@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -107,16 +107,16 @@ void *client_thread(void *arg) {
 
 void client_main(CLI *c) {
     s_log(LOG_DEBUG, "Service %s started", c->opt->servname);
-    if(c->opt->option.remote && c->opt->option.program) {
-            /* connect and exec options specified together */
-            /* -> spawn a local program instead of stdio */
+    if(c->opt->option.program && c->opt->option.remote) {
+            /* exec and connect options specified together
+             * -> spawn a local program instead of stdio */
         while((c->local_rfd.fd=c->local_wfd.fd=connect_local(c))>=0) {
             client_run(c);
             if(!c->opt->option.retry)
                 break;
             sleep(1); /* FIXME: not a good idea in ucontext threading */
             str_stats();
-            if(service_options.next)
+            if(service_options.next) /* don't str_cleanup in inetd mode */
                 str_cleanup();
         }
     } else
@@ -127,6 +127,12 @@ void client_main(CLI *c) {
 static void client_run(CLI *c) {
     int error;
 
+#ifndef USE_FORK
+    enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
+    ++num_clients;
+    leave_critical_section(CRIT_CLIENTS);
+#endif
+
     c->remote_fd.fd=-1;
     c->fd=-1;
     c->ssl=NULL;
@@ -184,6 +190,7 @@ static void client_run(CLI *c) {
     /* otherwise it will be retrieved by the init process and ignored */
     if(c->opt->option.program) /* 'exec' specified */
         child_status(); /* null SIGCHLD handler was used */
+    s_log(LOG_DEBUG, "Service %s finished", c->opt->servname);
 #else
     enter_critical_section(CRIT_CLIENTS); /* for multi-cpu machines */
     s_log(LOG_DEBUG, "Service %s finished (%d left)", c->opt->servname,
@@ -231,8 +238,8 @@ static void init_local(CLI *c) {
     c->local_rfd.is_socket=1;
     c->local_wfd.is_socket=1; /* TODO: It's not always true */
     /* it's a socket: lets setup options */
-    if(set_socket_options(c->local_rfd.fd, 1)<0)
-        longjmp(c->err, 1);
+    if(set_socket_options(c->local_rfd.fd, 1))
+        s_log(LOG_WARNING, "Failed to set local socket options");
 #ifdef USE_LIBWRAP
     libwrap_auth(c, accepted_address);
 #endif /* USE_LIBWRAP */
@@ -254,13 +261,19 @@ static void init_remote(CLI *c) {
         c->bind_addr=NULL; /* don't bind */
 
     /* setup c->remote_fd, now */
-    c->remote_fd.fd=c->opt->option.program ?
-        connect_local(c) :
-        connect_remote(c);
+    if(c->opt->option.remote) { /* try remote first for exec+connect targets */
+        c->remote_fd.fd=connect_remote(c);
+    } else if(c->opt->option.program) { /* exec+connect uses local fd */
+        c->remote_fd.fd=connect_local(c);
+    } else {
+        s_log(LOG_ERR, "INTERNAL ERROR: No target for remote socket");
+        longjmp(c->err, 1);
+    }
+
     c->remote_fd.is_socket=1; /* always! */
     s_log(LOG_DEBUG, "Remote FD=%d initialized", c->remote_fd.fd);
-    if(set_socket_options(c->remote_fd.fd, 2)<0)
-        longjmp(c->err, 1);
+    if(set_socket_options(c->remote_fd.fd, 2))
+        s_log(LOG_WARNING, "Failed to set remote socket options");
 }
 
 static void init_ssl(CLI *c) {
@@ -504,7 +517,7 @@ static void transfer(CLI *c) {
         }
 
         /****************************** check for errors on sockets */
-        err=s_poll_error(c->fds, c->sock_rfd->fd);
+        err=s_poll_error(c->fds, c->sock_rfd);
         if(err) {
             s_log(LOG_NOTICE,
                 "Error detected on socket (read) file descriptor: %s (%d)",
@@ -512,7 +525,7 @@ static void transfer(CLI *c) {
             longjmp(c->err, 1);
         }
         if(c->sock_wfd->fd != c->sock_rfd->fd) { /* performance optimization */
-            err=s_poll_error(c->fds, c->sock_wfd->fd);
+            err=s_poll_error(c->fds, c->sock_wfd);
             if(err) {
                 s_log(LOG_NOTICE,
                     "Error detected on socket write file descriptor: %s (%d)",
@@ -520,7 +533,7 @@ static void transfer(CLI *c) {
                 longjmp(c->err, 1);
             }
         }
-        err=s_poll_error(c->fds, c->ssl_rfd->fd);
+        err=s_poll_error(c->fds, c->ssl_rfd);
         if(err) {
             s_log(LOG_NOTICE,
                 "Error detected on SSL (read) file descriptor: %s (%d)",
@@ -528,7 +541,7 @@ static void transfer(CLI *c) {
             longjmp(c->err, 1);
         }
         if(c->ssl_wfd->fd != c->ssl_rfd->fd) { /* performance optimization */
-            err=s_poll_error(c->fds, c->ssl_wfd->fd);
+            err=s_poll_error(c->fds, c->ssl_wfd);
             if(err) {
                 s_log(LOG_NOTICE,
                     "Error detected on SSL write file descriptor: %s (%d)",
@@ -550,37 +563,17 @@ static void transfer(CLI *c) {
                 "s_poll_wait returned %d, but no descriptor is ready", err);
             longjmp(c->err, 1);
         }
-        /* these checks should no longer be needed */
-        /* I'm going to remove them soon */
-        if(!sock_open_rd && sock_can_rd) {
-            err=get_socket_error(c->sock_rfd->fd);
-            if(err) { /* really an error? */
-                s_log(LOG_ERR, "INTERNAL ERROR: "
-                    "Closed socket ready to read: %s (%d)",
-                    s_strerror(err), err);
-                longjmp(c->err, 1);
-            }
-            if(c->ssl_ptr) { /* anything left to write */
-                s_log(LOG_ERR, "INTERNAL ERROR: "
-                    "Closed socket ready to read: sending reset");
-                longjmp(c->err, 1);
-            }
-            s_log(LOG_ERR, "INTERNAL ERROR: "
-                "Closed socket ready to read: write close");
-            sock_open_wr=0; /* no further write allowed */
-            shutdown(c->sock_wfd->fd, SHUT_WR); /* send TCP FIN */
-        }
 
-        /****************************** send SSL close_notify message */
+        /****************************** send SSL close_notify alert */
         if(shutdown_wants_read || shutdown_wants_write) {
-            num=SSL_shutdown(c->ssl); /* send close_notify */
+            num=SSL_shutdown(c->ssl); /* send close_notify alert */
             if(num<0) /* -1 - not completed */
                 err=SSL_get_error(c->ssl, num);
             else /* 0 or 1 - success */
                 err=SSL_ERROR_NONE;
             switch(err) {
             case SSL_ERROR_NONE: /* the shutdown was successfully completed */
-                s_log(LOG_INFO, "SSL_shutdown successfully sent close_notify");
+                s_log(LOG_INFO, "SSL_shutdown successfully sent close_notify alert");
                 shutdown_wants_read=shutdown_wants_write=0;
                 break;
             case SSL_ERROR_SYSCALL: /* socket error */
@@ -679,7 +672,8 @@ static void transfer(CLI *c) {
             case SSL_ERROR_SYSCALL:
                 if(num && parse_socket_error(c, "SSL_read"))
                     break; /* a non-critical error: retry */
-                /* EOF -> buggy peer: no close_notify */
+                /* EOF -> buggy (e.g. Microsoft) peer:
+                 * SSL socket closed without close_notify alert */
                 if(c->sock_ptr) {
                     s_log(LOG_ERR,
                         "SSL socket closed on SSL_read with %d unsent byte(s)",
@@ -689,7 +683,7 @@ static void transfer(CLI *c) {
                 s_log(LOG_DEBUG, "SSL socket closed on SSL_read");
                 SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify received */
+            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
                 s_log(LOG_DEBUG, "SSL closed on SSL_read");
                 if(SSL_version(c->ssl)==SSL2_VERSION)
                     SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
@@ -730,7 +724,8 @@ static void transfer(CLI *c) {
             case SSL_ERROR_SYSCALL: /* socket error */
                 if(num && parse_socket_error(c, "SSL_write"))
                     break; /* a non-critical error: retry */
-                /* EOF -> buggy peer: no close_notify */
+                /* EOF -> buggy (e.g. Microsoft) peer:
+                 * SSL socket closed without close_notify alert */
                 if(c->sock_ptr) {
                     s_log(LOG_ERR,
                         "SSL socket closed on SSL_write with %d unsent byte(s)",
@@ -740,7 +735,7 @@ static void transfer(CLI *c) {
                 s_log(LOG_DEBUG, "SSL socket closed on SSL_write");
                 SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
                 break;
-            case SSL_ERROR_ZERO_RETURN: /* close_notify received */
+            case SSL_ERROR_ZERO_RETURN: /* close_notify alert received */
                 s_log(LOG_DEBUG, "SSL closed on SSL_write");
                 if(SSL_version(c->ssl)==SSL2_VERSION)
                     SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
@@ -756,16 +751,27 @@ static void transfer(CLI *c) {
 
         /****************************** check write shutdown conditions */
         if(sock_open_wr && SSL_get_shutdown(c->ssl)&SSL_RECEIVED_SHUTDOWN && !c->ssl_ptr) {
-            s_log(LOG_DEBUG, "Sending socket write shutdown");
             sock_open_wr=0; /* no further write allowed */
-            shutdown(c->sock_wfd->fd, SHUT_WR); /* send TCP FIN */
+            if(!c->sock_wfd->is_socket) {
+                s_log(LOG_DEBUG, "Closing the socket file descriptor");
+                sock_open_rd=0; /* file descriptor is ready to be closed */
+            } else if(!shutdown(c->sock_wfd->fd, SHUT_WR)) { /* send TCP FIN */
+                s_log(LOG_DEBUG, "Sent socket write shutdown");
+            } else {
+                s_log(LOG_DEBUG, "Failed to send socket write shutdown");
+                sock_open_rd=0; /* file descriptor is ready to be closed */
+            }
         }
         if(!(SSL_get_shutdown(c->ssl)&SSL_SENT_SHUTDOWN) && !sock_open_rd && !c->sock_ptr) {
-            s_log(LOG_DEBUG, "Sending SSL write shutdown");
             if(SSL_version(c->ssl)!=SSL2_VERSION) { /* SSLv3, TLSv1 */
-                shutdown_wants_write=1; /* initiate close_notify */
-            } else { /* no alerts in SSLv2 including close_notify alert */
-                shutdown(c->sock_rfd->fd, SHUT_RDWR); /* notify the kernel */
+                s_log(LOG_DEBUG, "Sending close_notify alert");
+                shutdown_wants_write=1;
+            } else { /* no alerts in SSLv2, including the close_notify alert */
+                s_log(LOG_DEBUG, "Closing SSLv2 socket");
+                if(c->ssl_rfd->is_socket)
+                    shutdown(c->ssl_rfd->fd, SHUT_RD); /* notify the kernel */
+                if(c->ssl_wfd->is_socket)
+                    shutdown(c->ssl_wfd->fd, SHUT_WR); /* send TCP FIN */
                 /* notify the OpenSSL library */
                 SSL_set_shutdown(c->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
             }
@@ -838,6 +844,9 @@ static int parse_socket_error(CLI *c, const char *text) {
 static void print_cipher(CLI *c) { /* print negotiated cipher */
     SSL_CIPHER *cipher;
     char *buf, *i, *j;
+#if !defined(OPENSSL_NO_COMP) && OPENSSL_VERSION_NUMBER>=0x0090800fL
+    const COMP_METHOD *compression, *expansion;
+#endif
 
     if(global_options.debug_level<LOG_INFO) /* performance optimization */
         return;
@@ -859,6 +868,14 @@ static void print_cipher(CLI *c) { /* print negotiated cipher */
     } while(*i++);
     s_log(LOG_INFO, "Negotiated ciphers: %s", buf);
     OPENSSL_free(buf);
+
+#if !defined(OPENSSL_NO_COMP) && OPENSSL_VERSION_NUMBER>=0x0090800fL
+    compression=SSL_get_current_compression(c->ssl);
+    expansion=SSL_get_current_expansion(c->ssl);
+    s_log(LOG_INFO, "Compression: %s, expansion: %s",
+        compression ? SSL_COMP_get_name(compression) : "null",
+        expansion ? SSL_COMP_get_name(expansion) : "null");
+#endif
 }
 
 static void auth_user(CLI *c, char *accepted_address) {
