@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -66,13 +66,17 @@ typedef struct sockaddr_list {                          /* list of addresses */
     u16 num;                                  /* how many addresses are used */
 } SOCKADDR_LIST;
 
+#ifndef OPENSSL_NO_COMP
 typedef enum {
-    COMP_NONE, COMP_ZLIB, COMP_RLE
+    COMP_NONE, COMP_DEFLATE, COMP_ZLIB, COMP_RLE
 } COMP_TYPE;
+#endif /* OPENSSL_NO_COMP */
 
 typedef struct {
         /* some data for SSL initialization in ssl.c */
+#ifndef OPENSSL_NO_COMP
     COMP_TYPE compression;                               /* compression type */
+#endif /* OPENSSL_NO_COMP */
     char *egd_sock;                       /* entropy gathering daemon socket */
     char *rand_file;                                /* file with random data */
     int random_bytes;                       /* how many random bytes to read */
@@ -267,6 +271,13 @@ typedef struct disk_file {
     /* the inteface is prepared to easily implement buffering if needed */
 } DISK_FILE;
 
+    /* FD definition for client.c */
+
+typedef struct {
+    int fd; /* file descriptor */
+    int is_socket; /* file descriptor is a socket */
+} FD;
+
 /**************************************** prototypes for stunnel.c */
 
 extern volatile int num_clients;
@@ -350,7 +361,7 @@ void s_poll_init(s_poll_set *);
 void s_poll_add(s_poll_set *, int, int, int);
 int s_poll_canread(s_poll_set *, int);
 int s_poll_canwrite(s_poll_set *, int);
-int s_poll_error(s_poll_set *, int);
+int s_poll_error(s_poll_set *, FD *);
 int s_poll_wait(s_poll_set *, int, int);
 
 #ifdef USE_WIN32
@@ -369,11 +380,6 @@ int make_sockets(int [2]);
 
 /**************************************** prototypes for client.c */
 
-typedef struct {
-    int fd; /* file descriptor */
-    int is_socket; /* file descriptor is a socket */
-} FD;
-
 typedef struct {
     SSL *ssl; /* SSL connnection */
     SERVICE_OPTIONS *opt;
@@ -461,13 +467,13 @@ int getnameinfo(const struct sockaddr *, int, char *, int, char *, int, int);
 /**************************************** prototypes for sthreads.c */
 
 typedef enum {
-    CRIT_CLIENTS, CRIT_SESSION, /* client.c */
-    CRIT_SSL,                   /* client.c */
-    CRIT_INET,                  /* resolver.c */
+    CRIT_CLIENTS, CRIT_SESSION, CRIT_SSL,   /* client.c */
+    CRIT_INET,                              /* resolver.c */
 #ifndef USE_WIN32
-    CRIT_LIBWRAP,               /* libwrap.c */
+    CRIT_LIBWRAP,                           /* libwrap.c */
 #endif
-    CRIT_SECTIONS               /* number of critical sections */
+    CRIT_LOG,                               /* log.c */
+    CRIT_SECTIONS                           /* number of critical sections */
 } SECTION_CODE;
 
 void enter_critical_section(SECTION_CODE);
@@ -553,7 +559,7 @@ void *str_realloc_debug(void *, size_t, char *, int);
 void str_detach_debug(void *, char *, int);
 #define str_detach(a) str_detach_debug((a), __FILE__, __LINE__)
 void str_free_debug(void *, char *, int);
-#define str_free(a) str_free_debug((a), __FILE__, __LINE__)
+#define str_free(a) str_free_debug((a), __FILE__, __LINE__), (a)=NULL
 char *str_dup(const char *);
 char *str_vprintf(const char *, va_list);
 char *str_printf(const char *, ...)
