@@ -1,6 +1,6 @@
 /*
  *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2011 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2012 Michal Trojnara <Michal.Trojnara@mirt.net>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -160,42 +160,45 @@ static int verify_callback(int preverify_ok, X509_STORE_CTX *callback_ctx) {
         /* our verify callback function */
     SSL *ssl;
     CLI *c;
+    X509 *cert;
+    int depth;
     char *subject_name;
 
     /* retrieve application specific data */
     ssl=X509_STORE_CTX_get_ex_data(callback_ctx,
         SSL_get_ex_data_X509_STORE_CTX_idx());
     c=SSL_get_ex_data(ssl, cli_index);
+    cert=X509_STORE_CTX_get_current_cert(callback_ctx);
+    depth=X509_STORE_CTX_get_error_depth(callback_ctx);
 
     /* certificate name for logging */
-    subject_name=X509_NAME_oneline(
-        X509_get_subject_name(callback_ctx->current_cert), NULL, 0);
+    subject_name=X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);
 
     s_log(LOG_DEBUG, "Starting certificate verification: depth=%d, %s",
-        callback_ctx->error_depth, subject_name);
+        depth, subject_name);
     if(!cert_check(c, callback_ctx, preverify_ok)) {
         s_log(LOG_WARNING, "Certificate check failed: depth=%d, %s",
-            callback_ctx->error_depth, subject_name);
+            depth, subject_name);
         OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
     if(!crl_check(c, callback_ctx)) {
         s_log(LOG_WARNING, "CRL check failed: depth=%d, %s",
-            callback_ctx->error_depth, subject_name);
+            depth, subject_name);
         OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
 #ifdef HAVE_OSSL_OCSP_H
     if(c->opt->option.ocsp && !ocsp_check(c, callback_ctx)) {
         s_log(LOG_WARNING, "OCSP check failed: depth=%d, %s",
-            callback_ctx->error_depth, subject_name);
+            depth, subject_name);
         OPENSSL_free(subject_name);
         return 0; /* reject connection */
     }
 #endif /* HAVE_OSSL_OCSP_H */
     /* errnum=X509_STORE_CTX_get_error(ctx); */
     s_log(LOG_NOTICE, "Certificate accepted: depth=%d, %s",
-        callback_ctx->error_depth, subject_name);
+        depth, subject_name);
     OPENSSL_free(subject_name);
     return 1; /* accept connection */
 }
@@ -207,30 +210,36 @@ static int cert_check(CLI *c, X509_STORE_CTX *callback_ctx, int preverify_ok) {
 #if OPENSSL_VERSION_NUMBER>=0x0090700fL
     ASN1_BIT_STRING *local_key, *peer_key;
 #endif
+    X509 *cert;
+    int depth;
 
     if(c->opt->verify_level<1) {
         s_log(LOG_INFO, "CERT: Verification not enabled");
         return 1; /* accept connection */
     }
+    cert=X509_STORE_CTX_get_current_cert(callback_ctx);
+    depth=X509_STORE_CTX_get_error_depth(callback_ctx);
     if(!preverify_ok) {
         /* remote site specified a certificate, but it's not correct */
-        if(c->opt->verify_level>=4 && callback_ctx->error_depth>0) {
+        if(c->opt->verify_level>=4 && depth>0) {
             s_log(LOG_INFO, "CERT: Invalid CA certificate ignored");
             return 1; /* accept connection */
         } else {    
             s_log(LOG_WARNING, "CERT: Verification error: %s",
-                X509_verify_cert_error_string(callback_ctx->error));
+                X509_verify_cert_error_string(
+                    X509_STORE_CTX_get_error(callback_ctx)));
             return 0; /* reject connection */
         }
     }
-    if(c->opt->verify_level>=3 && callback_ctx->error_depth==0) {
+    if(c->opt->verify_level>=3 && depth==0) {
         if(X509_STORE_get_by_subject(callback_ctx, X509_LU_X509,
-                X509_get_subject_name(callback_ctx->current_cert), &obj)!=1) {
-            s_log(LOG_WARNING, "CERT: Certificate not found in local repository");
+                X509_get_subject_name(cert), &obj)!=1) {
+            s_log(LOG_WARNING,
+                "CERT: Certificate not found in local repository");
             return 0; /* reject connection */
         }
 #if OPENSSL_VERSION_NUMBER>=0x0090700fL
-        peer_key=X509_get0_pubkey_bitstr(callback_ctx->current_cert);
+        peer_key=X509_get0_pubkey_bitstr(cert);
         local_key=X509_get0_pubkey_bitstr(obj.data.x509);
         if(!peer_key || !local_key || peer_key->length!=local_key->length ||
                 memcmp(peer_key->data, local_key->data, local_key->length)) {
